from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool

from .abc import TelegramType


@final
class ChatPermissions(TelegramType):
    """Describes actions that a non-administrator user is allowed to take in
    a chat.

    See: https://core.telegram.org/bots/api#chatpermissions
    """

    can_send_messages: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send text messages,
    contacts, invoices, locations and venues.
    """

    can_send_audios: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send audios."""

    can_send_documents: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send documents."""

    can_send_photos: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send photos."""

    can_send_videos: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send videos."""

    can_send_video_notes: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send video notes."""

    can_send_voice_notes: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send voice notes."""

    can_send_polls: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send polls."""

    can_send_other_messages: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to send animations, games,
    stickers and use inline bots.
    """

    can_add_web_page_previews: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to add web page previews to
    their messages.
    """

    can_change_info: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to change the chat title,
    photo and other settings.

    .. note::

        Ignored in public supergroups.
    """

    can_invite_users: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to invite new users to the chat."""  # noqa: E501

    can_pin_messages: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to pin messages.

    .. note::

        Ignored in public supergroups.
    """

    can_manage_topics: NoneBool = None
    """*Optional.* ``True``, if the user is allowed to create forum topics.

    .. note::

        If omitted defaults to the value of ``can_pin_messages``.
    """



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .user import User


@final
class VideoChatParticipantsInvited(TelegramType):
    """This object represents a service message about new members invited to a
    video chat.

    See: https://core.telegram.org/bots/api#videochatparticipantsinvited
    """

    users: list[User]
    """New members that were invited to the video chat."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt


@final
class SetGameScore(TelegramMethod[Message | bool]):
    """See :meth:`yatbaf.bot.Bot.set_game_score`"""

    user_id: int
    score: int
    force: NoneBool = None
    disable_edit_message: NoneBool = None
    chat_id: NoneInt = None
    message_id: NoneInt = None
    inline_message_id: NoneInt = None



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .user import User

# from datetime import datetime


@final
class ChatInviteLink(TelegramType):
    """Represents an invite link for a chat.

    See: https://core.telegram.org/bots/api#chatinvitelink
    """

    invite_link: str
    """The invite link. If the link was created by another chat administrator,
    then the second part of the link will be replaced with “…”.
    """

    creator: User
    """Creator of the link."""

    creates_join_request: bool
    """``True``, if users joining the chat via the link need to be approved by
    chat administrators.
    """

    is_primary: bool
    """``True``, if the link is primary."""

    is_revoked: bool
    """``True``, if the link is revoked."""

    name: NoneStr = None
    """*Optional.* Invite link name."""

    expire_date: NoneInt = None  # TODO: datetime
    """*Optional.* Point in time (Unix timestamp) when the link will expire or
    has been expired."""

    member_limit: NoneInt = None
    """*Optional.* The maximum number of users that can be members of the chat
    simultaneously after joining the chat via this invite link; 1-99999.
    """

    pending_join_request_count: NoneInt = None
    """*Optional.* Number of pending join requests created using this link."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneStr

from .abc import TelegramType


@final
class SwitchInlineQueryChosenChat(TelegramType):
    """This object represents an inline button that switches the current user
    to inline mode in a chosen chat, with an optional default inline query.

    See: https://core.telegram.org/bots/api#switchinlinequerychosenchat
    """

    query: NoneStr = None
    """*Optional.* The default inline query to be inserted in the input field.
    If left empty, only the bot's username will be inserted.
    """

    allow_user_chats: NoneBool = None
    """*Optional.* ``True``, if private chats with users can be chosen."""

    allow_bot_chats: NoneBool = None
    """*Optional.* ``True``, if private chats with bots can be chosen."""

    allow_group_chats: NoneBool = None
    """*Optional.* ``True``, if group and supergroup chats can be chosen."""

    allow_channel_chats: NoneBool = None
    """*Optional.* ``True``, if channel chats can be chosen."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class UserProfilePhotos(TelegramType):
    """This object represent a user's profile pictures.

    See: https://core.telegram.org/bots/api#userprofilephotos
    """

    total_count: int
    """Total number of profile pictures the target user has."""

    photos: list[list[PhotoSize]]
    """Requested profile pictures (in up to 4 sizes each)."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import ChatInviteLink

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class EditChatInviteLink(TelegramMethod[ChatInviteLink]):
    """See :meth:`yatbaf.bot.Bot.edit_chat_invite_link`"""

    chat_id: str | int
    invite_link: str
    name: NoneStr = None
    expire_date: NoneInt = None
    member_limit: NoneInt = None
    creates_join_request: NoneBool = None



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.enums import PollType
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup


@final
class SendPoll(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_poll`"""

    chat_id: str | int
    question: str
    options: list[str]
    message_thread_id: NoneInt = None
    is_anonymous: NoneBool = None
    type: PollType | None = None
    allows_multiple_answers: NoneBool = None
    correct_option_id: NoneInt = None
    explanation: NoneStr = None
    explanation_parse_mode: ParseMode | None = None
    explanation_entities: list[MessageEntity] | None = None
    open_period: NoneInt = None
    close_date: int | None = None  # timedelta
    is_closed: NoneBool = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Command
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


@pytest.fixture
def filter():
    return Command("foo")


@pytest.fixture
def handler(handler_func):
    return Handler(handler_func, UpdateType.MESSAGE)


@pytest.fixture
def router():
    return OnMessage()



from yatbaf.filters import And
from yatbaf.filters import Channel
from yatbaf.filters import Command
from yatbaf.filters import Content
from yatbaf.filters import Not
from yatbaf.filters import Or
from yatbaf.filters import User


class FalseFilter:
    priority = 100

    def check(self, _):
        return False


class TrueFilter:
    priority = 100

    def check(self, _):
        return True


def test_not_true(message):
    assert Not(FalseFilter()).check(message)


def test_not_false(message):
    assert not Not(TrueFilter()).check(message)


def test_or_true(message):
    assert Or(FalseFilter(), TrueFilter()).check(message)


def test_or_false(message):
    assert not Or(FalseFilter(), FalseFilter()).check(message)


def test_and_true(message):
    assert And(TrueFilter(), TrueFilter()).check(message)


def test_and_false(message):
    assert not And(TrueFilter(), FalseFilter()).check(message)


def test_priority():
    user = User(123)
    channel = Channel()
    command = Command("start")
    content = Content("text")

    assert Not(channel).priority == channel.priority
    assert And(command, user).priority == user.priority
    assert Or(channel, content).priority == channel.priority



from __future__ import annotations

__all__ = ("Text",)

from re import compile as re_compile
from typing import TYPE_CHECKING
from typing import final

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from re import Pattern

    from yatbaf.types import Message


@final
class Text:
    """Message text filter.

    See :class:`yatbaf.filters.content.Content`.

    Use it to filter message by text content::

        @on_message(filters=[Text(match="foo bar")])
        async def handler(message: Message) -> None:
            assert message.text == "foo bar"


        @on_message(filters=[Text(startswith="foo")])
        async def handler(message: Message) -> None:
            assert message.text.startswith("foo")


        @on_message(
            filters=[
                Text(
                    startswith="foo",
                    endswith="bar",
                    contains="baz",
                )
            ]
        )
        async def handler(message: Message) -> None:
            assert (
                (t := message.text).startswith("foo")
                and t.endswith("bar")
                and "baz" in t
            )

    """

    __slots__ = (
        "priority",
        "params",
        "ignore_case",
        "any_",
    )

    def __init__(
        self,
        startswith: str | None = None,
        endswith: str | None = None,
        match: str | None = None,
        contains: str | Iterable[str] | None = None,
        regexp: str | Pattern | None = None,
        any_: bool = False,
        ignore_case: bool = False,
        priority: int = 300,  # same as Content filter
    ) -> None:
        """
        :param startswith: *Optional.* Text prefix.
        :param endswith: *Optional.* Text suffix.
        :param match: *Optional.* Check if text is equal to passed string.
        :param contains: *Optional.* A string or list of strings to check if
            it is in the message.
        :param regexp: *Optional.* Regular expression.
        :param any_: Pass ``True``, if one match is enough. Default ``False``.
        :param ignore_case: Perform case-insensitive matching.
            Default ``False``.
        :param priority: Filter priority. Default 300.
        :raise ValueError: If all optional parameters is ``None``.

        .. note::

            You **must** pass at least one optional parameter, otherwise
            ``ValueError`` is raised.

        """
        params: list[Callable[[str], bool]] = []
        if startswith is not None:
            params.append(
                lambda s: s.startswith(startswith)  # type: ignore[arg-type]
            )

        if endswith is not None:
            params.append(
                lambda s: s.endswith(endswith)  # type: ignore[arg-type]
            )

        if match is not None:
            params.append(lambda s: s == match)

        if contains is not None:
            if isinstance(contains, str):
                contains = (contains,)
            # yapf: disable
            params.append(
                lambda s: any(w in s for w in contains)  # type: ignore[union-attr]  # noqa: E501
            )
            # yapf: enable

        if regexp is not None:
            if isinstance(regexp, str | bytes):
                regexp = re_compile(regexp)
            # re.search returns re.Match or None, re.Match is always True.
            params.append(regexp.search)  # type: ignore[arg-type]

        if not params:
            raise ValueError(
                "No parameters. "
                "You must pass at least one optional parameter."
            )

        self.ignore_case = ignore_case
        self.params = params
        self.priority = priority
        self.any_ = any_

    def check(self, update: Message) -> bool:
        # same behavior as Content(TEXT) filter
        if (text := update.text) is None:
            return False

        f = any if self.any_ else all
        text = text.lower() if self.ignore_case else text
        return f(p(text) for p in self.params)



import pytest

from yatbaf.helpers import parse_command_args


@pytest.mark.parametrize(
    "text,result",
    [
        ("/foo", []),
        ("/foo bar", ["bar"]),
        ("/foo@bot bar baz", ["bar", "baz"]),
        ("/foo Bar baZ 123", ["Bar", "baZ", "123"]),
    ],
)
def test_parse_args(text, result):
    assert parse_command_args(text) == result



__all__ = (
    "Bot",
    "LongPolling",
    "parse_command",
    "OnCallbackQuery",
    "OnChannelPost",
    "OnChatJoinRequest",
    "OnChatMemeber",
    "OnChosenInlineResult",
    "OnEditedChannelPost",
    "OnEditedMessage",
    "OnMessageReaction",
    "OnMessageReactionCount",
    "OnInlineQuery",
    "OnMessage",
    "OnMyChatMember",
    "OnPoll",
    "OnPollAnswer",
    "OnPreCheckoutQuery",
    "OnShippingQuery",
    "OnChatBoost",
    "OnRemovedChatBoost",
    "on_message",
    "on_edited_message",
    "on_message_reaction",
    "on_message_reaction_count",
    "on_channel_post",
    "on_edited_channel_post",
    "on_inline_query",
    "on_chosen_inline_result",
    "on_callback_query",
    "on_shipping_query",
    "on_pre_checkout_query",
    "on_poll",
    "on_poll_answer",
    "on_my_chat_member",
    "on_chat_member",
    "on_chat_join_request",
    "on_chat_boost",
    "on_removed_chat_boost",
    "Handler",
)

import logging

from .bot import Bot
from .handler import Handler
from .handler import on_callback_query
from .handler import on_channel_post
from .handler import on_chat_boost
from .handler import on_chat_join_request
from .handler import on_chat_member
from .handler import on_chosen_inline_result
from .handler import on_edited_channel_post
from .handler import on_edited_message
from .handler import on_inline_query
from .handler import on_message
from .handler import on_message_reaction
from .handler import on_message_reaction_count
from .handler import on_my_chat_member
from .handler import on_poll
from .handler import on_poll_answer
from .handler import on_pre_checkout_query
from .handler import on_removed_chat_boost
from .handler import on_shipping_query
from .long_polling import LongPolling
from .router import OnCallbackQuery
from .router import OnChannelPost
from .router import OnChatBoost
from .router import OnChatJoinRequest
from .router import OnChatMemeber
from .router import OnChosenInlineResult
from .router import OnEditedChannelPost
from .router import OnEditedMessage
from .router import OnInlineQuery
from .router import OnMessage
from .router import OnMessageReaction
from .router import OnMessageReactionCount
from .router import OnMyChatMember
from .router import OnPoll
from .router import OnPollAnswer
from .router import OnPreCheckoutQuery
from .router import OnRemovedChatBoost
from .router import OnShippingQuery
from .utils import parse_command

logging.getLogger(__name__).addHandler(logging.NullHandler())
del logging



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool

# from datetime import timedelta


@final
class BanChatMember(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.ban_chat_member`"""

    chat_id: str | int
    user_id: int
    until_date: int | None = None  # TODO: timedelta
    revoke_messages: NoneBool = None



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramType
from .mask_position import MaskPosition

if TYPE_CHECKING:
    from yatbaf.typing import InputFile


@final
class InputSticker(TelegramType):
    """This object describes a sticker to be added to a sticker set.

    See: https://core.telegram.org/bots/api#inputsticker
    """

    sticker: InputFile | str
    """The added sticker."""

    emoji_list: list[str]
    """List of 1-20 emoji associated with the sticker"""

    mask_position: MaskPosition | None = None
    """*Optional.* Position where the mask should be placed on faces. For
    “mask” stickers only.
    """

    keywords: list[str] | None = None
    """*Optional.* List of 0-20 search keywords for the sticker with total
    length of up to 64 characters. For “regular” and “custom_emoji” stickers
    only.
    """

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("sticker",)



from __future__ import annotations

from typing import final

from yatbaf.types import MenuButton

from .abc import TelegramMethod


@final
class GetChatMenuButton(TelegramMethod[MenuButton]):
    """See :meth:`yatbaf.bot.Bot.get_chat_menu_button`"""

    chat_id: str | int



from __future__ import annotations

from typing import final

from yatbaf.types import Sticker

from .abc import TelegramMethod


@final
class GetCustomEmojiStickers(TelegramMethod[list[Sticker]]):
    """See :meth:`yatbaf.bot.Bot.get_custom_emoji_stickers`"""

    custom_emoji_ids: list[str]



import pytest

from yatbaf import types


@pytest.fixture
def api_types(monkeypatch, api_spec):
    types = api_spec["types"]
    monkeypatch.delitem(types, "InputFile")
    return types


def test_type_object(api_types):
    for api_type in api_types.values():
        assert hasattr(types, api_type["name"])


def test_type_fields(api_types):
    for api_type in api_types.values():
        type_obj = getattr(types, api_type["name"])
        for field in api_type.get("fields", []):
            field_name = n if (n := field["name"]) != "from" else "from_"
            assert hasattr(type_obj, field_name)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .chat import Chat
from .user import User


@final
class GiveawayWinners(TelegramType):
    """This object represents a message about the completion of a giveaway with
    public winners.

    See: https://core.telegram.org/bots/api#giveawaywinners
    """

    chat: Chat
    """The chat that created the giveaway."""

    giveaway_message_id: int
    """Identifier of the messsage with the giveaway in the chat."""

    winners_selection_date: int
    """Point in time (Unix timestamp) when winners of the giveaway were selected."""  # noqa: E501

    winner_count: int
    """Total number of winners in the giveaway."""

    winners: list[User]
    """List of up to 100 winners of the giveaway."""

    additional_chat_count: NoneInt = None
    """*Optional.* The number of other chats the user had to join in order to be
    eligible for the giveaway.
    """

    premium_subscription_month_count: NoneInt = None
    """*Optional.* The number of months the Telegram Premium subscription won
    from the giveaway will be active for.
    """

    unclaimed_prize_count: NoneInt = None
    """*Optional.* Number of undistributed prizes."""

    only_new_members: NoneBool = None
    """*Optional.* ``True``, if only users who had joined the chats after the
    giveaway started were eligible to win.
    """

    was_refunded: NoneBool = None
    """*Optional.* ``True``, if the giveaway was canceled because the payment
    for it was refunded.
    """

    prize_description: NoneStr = None
    """*Optional.* Description of additional giveaway prize."""



from __future__ import annotations

__all__ = (
    "InputMedia",
    "InputMediaAnimation",
    "InputMediaAudio",
    "InputMediaDocument",
    "InputMediaPhoto",
    "InputMediaVideo",
)

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Literal
from typing import TypeAlias
from typing import final

from msgspec import field

from .abc import TelegramType

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.typing import InputFile
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr

    from .message_entity import MessageEntity


@final
class InputMediaPhoto(TelegramType):
    """Represents a photo to be sent.

    See: https://core.telegram.org/bots/api#inputmediaphoto
    """

    media: str | InputFile
    """File to send."""

    caption: NoneStr = None
    """*Optional.* Caption of the photo to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the photo caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    has_spoiler: NoneBool = None
    """*Optional.* Pass ``True`` if the photo needs to be covered with a
    spoiler animation.
    """

    type: Literal["photo"] = field(default_factory=lambda: "photo")
    """Type of the result, must be photo."""

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("media",)


@final
class InputMediaAudio(TelegramType):
    """Represents an audio file to be treated as music to be sent.

    See: https://core.telegram.org/bots/api#inputmediaaudio
    """

    media: str | InputFile
    """File to send."""

    thumbnail: InputFile | str | None = None
    """Optional. Thumbnail of the file sent."""

    caption: NoneStr = None
    """*Optional.* Caption of the audio to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the audio caption. See
    formatting options for more details.
    """

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    duration: NoneInt = None
    """*Optional.* Duration of the audio in seconds."""

    performer: NoneStr = None
    """*Optional.* Performer of the audio."""

    title: NoneStr = None
    """*Optional.* Title of the audio."""

    type: Literal["audio"] = field(default_factory=lambda: "audio")
    """Type of the result, must be audio."""

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("media", "thumbnail")


@final
class InputMediaVideo(TelegramType):
    """Represents a video to be sent.

    See: https://core.telegram.org/bots/api#inputmediavideo
    """

    media: str | InputFile
    """File to send."""

    thumbnail: InputFile | str | None = None
    """*Optional.* Thumbnail of the file sent; can be ignored if thumbnail
    generation for the file is supported server-side. The thumbnail should be
    in JPEG format and less than 200 kB in size. A thumbnail's width and height
    should not exceed 320.
    """

    caption: NoneStr = None
    """*Optional.* Caption of the video to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the video caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    width: NoneInt = None
    """*Optional.* Video width."""

    height: NoneInt = None
    """*Optional.* Video height."""

    duration: NoneInt = None
    """*Optional.* Video duration in seconds."""

    supports_streaming: NoneBool = None
    """
    *Optional.* Pass ``True`` if the uploaded video is suitable for streaming.
    """

    has_spoiler: NoneBool = None
    """*Optional.* Pass ``True`` if the video needs to be covered with a
    spoiler animation.
    """

    type: Literal["video"] = field(default_factory=lambda: "video")
    """Type of the result, must be video."""

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("media", "thumbnail")


@final
class InputMediaDocument(TelegramType):
    """Represents a general file to be sent.

    See: https://core.telegram.org/bots/api#inputmediadocument
    """

    media: str | InputFile
    """File to send."""

    thumbnail: InputFile | str | None = None
    """*Optional.* Thumbnail of the file sent; can be ignored if thumbnail
    generation for the file is supported server-side. The thumbnail should be
    in JPEG format and less than 200 kB in size. A thumbnail's width and height
    should not exceed 320.
    """

    caption: NoneStr = None
    """*Optional.* Caption of the document to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the document caption."""

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    disable_content_type_detection: NoneBool = None
    """*Optional.* Disables automatic server-side content type detection for
    files uploaded using multipart/form-data. Always ``True``, if the document
    is sent as part of an album.
    """

    type: Literal["document"] = field(default_factory=lambda: "document")
    """Type of the result, must be document."""

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("media", "thumbnail")


@final
class InputMediaAnimation(TelegramType):
    """Represents an animation file (GIF or H.264/MPEG-4 AVC video without
    sound) to be sent.

    See: https://core.telegram.org/bots/api#inputmediaanimation
    """

    media: str | InputFile
    """File to send."""

    thumbnail: InputFile | str | None = None
    """*Optional.* Thumbnail of the file sent."""

    caption: NoneStr = None
    """*Optional.* Caption of the animation to be sent, 0-1024 characters after
    entities parsing.
    """

    parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the animation caption. See
    formatting options for more details.
    """

    caption_entities: list[MessageEntity] | None = None
    """*Optional.* List of special entities that appear in the caption, which
    can be specified instead of ``parse_mode``.
    """

    width: NoneInt = None
    """*Optional.* Animation width."""

    height: NoneInt = None
    """*Optional.* Animation height."""

    duration: NoneInt = None
    """*Optional.* Animation duration in seconds."""

    has_spoiler: NoneBool = None
    """*Optional.* Pass ``True`` if the animation needs to be covered with a
    spoiler animation.
    """

    type: Literal["animation"] = field(default_factory=lambda: "animation")
    """Type of the result, must be animation."""

    __type_file_fields__: ClassVar[tuple[str, ...]] = ("media", "thumbnail")


# https://core.telegram.org/bots/api#inputmedia
InputMedia: TypeAlias = (
    "InputMediaAnimation "
    "| InputMediaAudio "
    "| InputMediaDocument "
    "| InputMediaPhoto "
    "| InputMediaVideo"
)



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat_boost_source import ChatBoostSource


@final
class ChatBoost(TelegramType):
    """This object contains information about a chat boost.

    See: https://core.telegram.org/bots/api#chatboost
    """

    boost_id: str
    """Unique identifier of the boost."""

    add_date: int
    """Point in time (Unix timestamp) when the chat was boosted."""

    expiration_date: int
    """Point in time (Unix timestamp) when the boost will automatically expire,
    unless the booster's Telegram Premium subscription is prolonged.
    """

    source: ChatBoostSource
    """Source of the added boost."""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class ChatPhoto(TelegramType):
    """This object represents a chat photo.

    See: https://core.telegram.org/bots/api#chatphoto
    """

    small_file_id: str
    """File identifier of small (160x160) chat photo. This file_id can be used
    only for photo download and only for as long as the photo is not changed.
    """

    small_file_unique_id: str
    """Unique file identifier of small (160x160) chat photo, which is supposed
    to be the same over time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """

    big_file_id: str
    """File identifier of big (640x640) chat photo. This ``file_id`` can be used
    only for photo download and only for as long as the photo is not changed.
    """

    big_file_unique_id: str
    """Unique file identifier of big (640x640) chat photo, which is supposed to
    be the same over time and for different bots.

    .. warning::

        Can't be used to download or reuse the file.
    """



from yatbaf.middleware import Middleware


def test_middleware():

    class SimpleMiddleware:

        def __init__(self, next_):
            self.next_ = next_

    middleware = Middleware(SimpleMiddleware)
    assert not middleware.is_local
    assert not middleware.is_handler

    result = middleware(next_ := object())
    assert result.next_ is next_


def test_middleware_params():

    class ParamsMiddleware:

        def __init__(self, next_, param1, param2):
            self.next_ = next_
            self.param1 = param1
            self.param2 = param2

    middleware = Middleware(
        ParamsMiddleware,
        param1=1,
        param2=2,
    )

    result = middleware(next_ := object())
    assert result.next_ is next_
    assert result.param1 == 1
    assert result.param2 == 2



from __future__ import annotations

from typing import final

from yatbaf.enums import StickerType

from .abc import TelegramType
from .photo_size import PhotoSize
from .sticker import Sticker


@final
class StickerSet(TelegramType):
    """This object represents a sticker set.

    See: https://core.telegram.org/bots/api#stickerset
    """

    name: str
    """Sticker set name."""

    title: str
    """Sticker set title."""

    sticker_type: StickerType
    """Type of stickers in the set."""

    is_animated: bool
    """True, if the sticker set contains animated stickers."""

    is_video: bool
    """True, if the sticker set contains video stickers."""

    stickers: list[Sticker]
    """List of all set stickers."""

    thumbnail: PhotoSize
    """Optional. Sticker set thumbnail in the .WEBP, .TGS, or .WEBM format."""



from __future__ import annotations

from typing import final

from yatbaf.types import UserChatBoosts

from .abc import TelegramMethod


@final
class GetUserChatBoosts(TelegramMethod[UserChatBoosts]):
    """See :meth:`~yatbaf.bot.Bot.get_user_chat_boosts`."""

    chat_id: int | str
    user_id: int



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.enums import UpdateType
    from yatbaf.typing import InputFile
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneStr


@final
class SetWebhook(TelegramMethodWithFile[bool]):
    """See :meth:`yatbaf.bot.Bot.set_webhook`"""

    url: str
    certificate: InputFile | None = None
    ip_address: NoneStr = None
    max_connections: int | None = None
    allowed_updates: list[UpdateType] | None = None
    drop_pending_updates: NoneBool = None
    secret_token: str | None = None

    __meth_file_fields__: ClassVar[tuple[str, ...]] = ("certificate",)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from .abc import TelegramMethodWithFile

if TYPE_CHECKING:
    from yatbaf.typing import InputFile


@final
class SetChatPhoto(TelegramMethodWithFile[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_photo`"""

    chat_id: str | int
    photo: InputFile

    __meth_file_fields__: ClassVar[tuple[str, ...]] = ("photo",)



from msgspec import json as jsonlib

from yatbaf.enums import ParseMode
from yatbaf.methods import EditMessageMedia
from yatbaf.methods import GetMe
from yatbaf.methods import SendMediaGroup
from yatbaf.methods import SendMessage
from yatbaf.methods import SendPhoto
from yatbaf.types import InputMediaDocument
from yatbaf.types import InputMediaPhoto


def test_encode_empty():
    method = GetMe()
    data, files = method._encode_params()
    assert data is None
    assert files is None


def test_encode_params():
    method = SendMessage(
        chat_id=12345,
        text="*foo bar*",
        disable_notification=True,
        parse_mode=ParseMode.MARKDOWN,
    )
    data, files = method._encode_params()
    assert isinstance(data, bytes)
    assert files is None
    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "text": "*foo bar*",
        "disable_notification": True,
        "parse_mode": "MarkdownV2",
    }


def test_encode_file_id():
    method = SendPhoto(chat_id=123, photo="fileid")
    data, files = method._encode_params()
    assert files is None
    assert jsonlib.decode(data) == {"chat_id": 123, "photo": "fileid"}


def test_encode_file_content():
    photo = object()
    method = SendPhoto(chat_id=123, photo=photo)
    data, files = method._encode_params()

    assert files is not None
    assert isinstance(data, dict)
    assert data == {"chat_id": 123, "photo": (fn := method.photo)}
    assert files[fn.split("//")[1]] is photo


def test_encode_media_file_content():
    photo = object()
    method = EditMessageMedia(
        chat_id=12345,
        message_id=12345,
        media=InputMediaPhoto(media=photo),
    )
    data, files = method._encode_params()
    media = jsonlib.decode(data["media"])

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "message_id": 12345,
        "media": jsonlib.encode({
            "media": (fn := media["media"]),
            "type": "photo",
        }).decode(),
    }
    assert files[fn.split("//")[1]] is photo


def test_encode_media_filed_id():
    method = EditMessageMedia(
        chat_id=12345,
        message_id=12345,
        media=InputMediaPhoto(media="fileid"),
    )
    data, files = method._encode_params()

    assert files is None
    assert isinstance(data, bytes)

    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "message_id": 12345,
        "media": {
            "media": "fileid",
            "type": "photo",
        }
    }


def test_encode_media_group_file_content():
    photo = object()
    method = SendMediaGroup(
        media=[InputMediaPhoto(media=photo)],
        chat_id=12345,
    )
    data, files = method._encode_params()
    media = jsonlib.decode(data["media"])

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": jsonlib.encode([
            {
                "media": (fn := media[0]["media"]),
                "type": "photo",
            },
        ]).decode(),
    }
    assert files[fn.split("//")[1]] is photo


def test_encode_media_group_file_content_1():
    doc1 = object()
    doc2 = object()
    doc2_th = object()
    doc3 = object()

    method = SendMediaGroup(
        media=[
            InputMediaDocument(media=doc1),
            InputMediaDocument(media=doc2, thumbnail=doc2_th),
            InputMediaDocument(media=doc3),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()
    media = jsonlib.decode(data["media"])

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": jsonlib.encode([
            {
                "media": (fn1 := media[0]["media"]),
                "type": "document",
            },
            {
                "media": (fn2 := media[1]["media"]),
                "thumbnail": (fn2_th := media[1]["thumbnail"]),
                "type": "document",
            },
            {
                "media": (fn3 := media[2]["media"]),
                "type": "document",
            },
        ]).decode(),
    }
    assert files[fn1.split("//")[1]] is doc1
    assert files[fn2.split("//")[1]] is doc2
    assert files[fn2_th.split("//")[1]] is doc2_th
    assert files[fn3.split("//")[1]] is doc3


def test_encode_media_group_file_id_content_mix():
    doc = object()
    doc_th = object()

    method = SendMediaGroup(
        media=[
            InputMediaDocument(media="fileid1"),
            InputMediaDocument(media="fileid2", thumbnail=doc_th),
            InputMediaDocument(media=doc),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()
    media = jsonlib.decode(data["media"])

    assert files is not None
    assert isinstance(data, dict)

    assert data == {
        "chat_id": 12345,
        "media": jsonlib.encode([
            {
                "media": "fileid1",
                "type": "document",
            },
            {
                "media": "fileid2",
                "thumbnail": (fn2_th := media[1]["thumbnail"]),
                "type": "document",
            },
            {
                "media": (fn3 := media[2]["media"]),
                "type": "document",
            },
        ]).decode(),
    }

    assert files[fn2_th.split("//")[1]] is doc_th
    assert files[fn3.split("//")[1]] is doc


def test_encode_media_group_file_id():
    method = SendMediaGroup(
        media=[
            InputMediaDocument(media="fileid1"),
            InputMediaDocument(media="fileid2", thumbnail="thid"),
            InputMediaDocument(media="fileid3"),
        ],
        chat_id=12345,
    )
    data, files = method._encode_params()

    assert files is None
    assert isinstance(data, bytes)
    assert jsonlib.decode(data) == {
        "chat_id": 12345,
        "media": [
            {
                "media": "fileid1",
                "type": "document",
            },
            {
                "media": "fileid2",
                "thumbnail": "thid",
                "type": "document",
            },
            {
                "media": "fileid3",
                "type": "document",
            },
        ],
    }



import httpx
import pytest
import pytest_asyncio
from pytest_httpx import IteratorStream

from yatbaf.exceptions import RequestTimeoutError
from yatbaf.http import HttpClient
from yatbaf.models import Request
from yatbaf.telegram import SERVER_URL


@pytest.fixture
def request_json(token) -> Request:
    method = "sendmessage"
    return Request(
        url=f"{SERVER_URL}/bot{token}/{method}",
        method=method,
        content=b'{"chat_id":123,"text":"hello"}',
        headers={"Content-Type": "application/json"},
    )


@pytest.fixture
def request_file(token) -> Request:
    method = "senddocument"
    return Request(
        url=f"{SERVER_URL}/bot{token}/{method}",
        method=method,
        data={"chat_id": 123},
        files={"k": ("file.txt", b"hello\nworld")},
    )


@pytest_asyncio.fixture
async def client() -> HttpClient:
    client = HttpClient()
    yield client
    await client.close()


@pytest.mark.asyncio
async def test_post(token, httpx_mock, request_json, client):
    content = b'{"ok":true}'

    httpx_mock.add_response(
        method="POST",
        url=f"{SERVER_URL}/bot{token}/{request_json.method}",
        content=content,
    )

    result = await client.send_post(request_json)

    assert result.status_code == 200
    assert result.content == content


@pytest.mark.asyncio
async def test_stream(token, httpx_mock, client):
    content = [b'chunk' + str(x).encode() for x in range(5)]
    file_id = "qwerty1234"
    url = f"{SERVER_URL}/file/bot{token}/{file_id}"

    httpx_mock.add_response(
        url=url,
        stream=IteratorStream(content),
    )

    async with client.file_stream(url, len(content[0])) as f:
        assert f.status_code == 200
        assert [x async for x in f.content] == content


@pytest.mark.asyncio
async def test_post_timeout(httpx_mock, client, request_json):
    exception = httpx.ReadTimeout("timeout")
    httpx_mock.add_exception(exception)

    with pytest.raises(RequestTimeoutError) as error:
        await client.send_post(request_json)

    assert error.value.orig is exception
    assert error.value.method == request_json.method


@pytest.mark.asyncio
async def test_stream_timeout(httpx_mock, client):
    exception = httpx.ReadTimeout("timeout")
    httpx_mock.add_exception(exception)

    url = f"{SERVER_URL}/file/bot1234:qwer/1234"
    with pytest.raises(RequestTimeoutError) as error:
        async with client.file_stream(url, 64):
            pass

    assert error.value.orig is exception
    assert error.value.method == "file"



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class ShippingAddress(TelegramType):
    """This object represents a shipping address.

    See: https://core.telegram.org/bots/api#shippingaddress
    """

    country_code: str
    """Two-letter ISO 3166-1 alpha-2 country code."""

    state: str
    """State, if applicable."""

    city: str
    """City."""

    street_line1: str
    """First line for the address."""

    street_line2: str
    """Second line for the address."""

    post_code: str
    """Address post code."""



from __future__ import annotations

from typing import final

from yatbaf.enums import ParseMode
from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .message_entity import MessageEntity


@final
class ReplyParameters(TelegramType):
    """Describes reply parameters for the message that is being sent.

    See: https://core.telegram.org/bots/api#replyparameters
    """

    message_id: int
    """Identifier of the message that will be replied to in the current chat,
    or in the chat ``chat_id`` if it is specified.
    """

    chat_id: str | int | None = None
    """*Optional.* If the message to be replied to is from a different chat,
    unique identifier for the chat or username of the channel (in the format
    @channelusername).
    """

    allow_sending_without_reply: NoneBool = None
    """*Optional.* Pass ``True`` if the message should be sent even if the
    specified message to be replied to is not found; can be used only for
    replies in the same chat and forum topic.
    """

    quote: NoneStr = None
    """*Optional.* Quoted part of the message to be replied to; 0-1024
    characters after entities parsing. The quote must be an exact substring of
    the message to be replied to, including *bold*, *italic*, *underline*,
    *strikethrough*, *spoiler*, and *custom_emoji* entities. The message will
    fail to send if the quote isn't found in the original message.
    """

    quote_parse_mode: ParseMode | None = None
    """*Optional.* Mode for parsing entities in the quote."""

    quote_entities: list[MessageEntity] | None = None
    """*Optional.* A list of special entities that appear in the quote. It can
    be specified instead of ``quote_parse_mode``.
    """

    quote_position: NoneInt = None
    """*Optional.* Position of the quote in the original message in UTF-16 code
    units.
    """



import pytest
import pytest_asyncio
from pytest_httpx import IteratorStream

from yatbaf.exceptions import ChatMigratedError
from yatbaf.exceptions import FileDownloadError
from yatbaf.exceptions import FloodError
from yatbaf.exceptions import InternalError
from yatbaf.exceptions import MethodInvokeError
from yatbaf.exceptions import TokenError
from yatbaf.exceptions import WebhookConflictError
from yatbaf.http import HttpClient
from yatbaf.methods import GetMe
from yatbaf.methods import SendDocument
from yatbaf.methods import SendMessage
from yatbaf.models import Request
from yatbaf.models import ResponseOk
from yatbaf.telegram import SERVER_URL
from yatbaf.telegram import TelegramClient
from yatbaf.types import File
from yatbaf.types import User

TOKEN = "123456789:totkentest"
URL = f"{SERVER_URL}/bot{TOKEN}"


@pytest_asyncio.fixture
async def tg() -> TelegramClient:
    tg_client = TelegramClient(
        token=TOKEN,
        client=HttpClient(),
    )
    yield tg_client
    await tg_client.close()


@pytest.fixture
def json_error():
    return {
        "ok": False,
        "description": "description",
        "error_code": 400,
    }


@pytest.mark.asyncio
async def test_prepare_request_empty():
    method = GetMe()
    req = await TelegramClient._prepare_request(URL, method)
    assert req == Request(
        url=f"{URL}/{method.__meth_name__}",
        content=None,
        method=str(method),
        data=None,
        files=None,
        headers=None,
    )


@pytest.mark.asyncio
async def test_prepare_request():
    method = SendMessage(text="ping", chat_id=12345)
    req = await TelegramClient._prepare_request(URL, method)
    assert req == Request(
        url=f"{URL}/{method.__meth_name__}",
        content=method._encode_params()[0],
        method=str(method),
        data=None,
        files=None,
        headers={"Content-Type": "application/json"},
    )


@pytest.mark.asyncio
async def test_prepare_request_file_content():

    class Doc:
        file_name = "file.txt"

        async def read(self):
            return b"content"

    document = Doc()

    method = SendDocument(document=document, chat_id=12345)
    req = await TelegramClient._prepare_request(URL, method)
    fn = method.document
    assert req == Request(
        url=f"{URL}/{method.__meth_name__}",
        content=None,
        method=str(method),
        data={
            "chat_id": 12345,
            "document": fn,
        },
        files={fn.split("//")[1]: ("file.txt", b"content")},
        headers=None,
    )


@pytest.mark.asyncio
async def test_prepare_request_file_id():
    document = "fileid"
    method = SendDocument(document=document, chat_id=12345)
    req = await TelegramClient._prepare_request(URL, method)
    assert req == Request(
        url=f"{URL}/{method.__meth_name__}",
        content=method._encode_params()[0],
        method=str(method),
        data=None,
        files=None,
        headers={"Content-Type": "application/json"},
    )


@pytest.mark.asyncio
async def test_invoke(httpx_mock, tg):
    json = {
        "ok": True,
        "result": {
            "id": 123456789,
            "is_bot": True,
            "first_name": "Test First Name",
        },
    }
    httpx_mock.add_response(method="POST", url=f"{URL}/getme", json=json)
    resp = await tg.invoke(GetMe())

    assert isinstance(resp, ResponseOk)
    assert isinstance(resp.result, User)
    assert resp.result.id == json["result"]["id"]
    assert resp.result.is_bot


@pytest.mark.asyncio
async def test_file_stream(httpx_mock, tg):
    content = b"hello world"
    httpx_mock.add_response(
        method="GET",
        url=f"{SERVER_URL}/file/bot{TOKEN}/123",
        stream=IteratorStream([content]),
    )
    gen = tg.download_file(
        File(file_id=123, file_unique_id="123", file_path="123"),
        chunk_size=64 * 1024,
    )
    file = b"".join([x async for x in gen])
    assert file == content


@pytest.mark.asyncio
async def test_token_error(httpx_mock, tg, json_error):
    status = 401
    json_error["error_code"] = status

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    with pytest.raises(TokenError):
        await tg.invoke(GetMe())


@pytest.mark.asyncio
async def test_webhook_error(httpx_mock, tg, json_error):
    status = 409
    json_error["error_code"] = status

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    with pytest.raises(WebhookConflictError):
        await tg.invoke(GetMe())


@pytest.mark.asyncio
async def test_flood_error(httpx_mock, tg, json_error):
    status = 429
    json_error["error_code"] = status
    json_error["parameters"] = {"retry_after": 123}

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    method = GetMe()
    with pytest.raises(FloodError) as error:
        await tg.invoke(method)

    assert error.value.retry_after == 123
    assert error.value.method is method


@pytest.mark.asyncio
async def test_chat_migrated_error(httpx_mock, tg, json_error):
    status = 400
    json_error["error_code"] = status
    json_error["parameters"] = {"migrate_to_chat_id": 123}

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    method = GetMe()
    with pytest.raises(ChatMigratedError) as error:
        await tg.invoke(method)

    assert error.value.migrate_to_chat_id == 123
    assert error.value.method is method


@pytest.mark.asyncio
async def test_file_error(httpx_mock, tg, json_error):
    status = 400
    json_error["error_code"] = status

    httpx_mock.add_response(
        method="GET",
        status_code=status,
        json=json_error,
    )

    with pytest.raises(FileDownloadError) as error:
        gen = tg.download_file(
            File(file_id=123, file_unique_id="123"),
            chunk_size=64 * 1024,
        )
        async for x in gen:
            pass

    assert error.value.error_code == status
    assert error.value.description == json_error["description"]


@pytest.mark.asyncio
async def test_internal_error(httpx_mock, tg, json_error):
    status = 500
    json_error["error_code"] = status

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    with pytest.raises(InternalError):
        await tg.invoke(GetMe())


@pytest.mark.asyncio
async def test_method_error(httpx_mock, tg, json_error):
    status = 400
    json_error["error_code"] = status

    httpx_mock.add_response(
        method="POST",
        status_code=status,
        json=json_error,
    )

    method = GetMe()
    with pytest.raises(MethodInvokeError) as error:
        await tg.invoke(method)

    assert error.value.method is method
    assert error.value.error_code == status
    assert error.value.description == json_error["description"]



from __future__ import annotations

from typing import final

from yatbaf.enums import MaskPositionPoint

from .abc import TelegramType


@final
class MaskPosition(TelegramType):
    """This object describes the position on faces where a mask should be placed
    by default.

    See: https://core.telegram.org/bots/api#maskposition
    """

    point: MaskPositionPoint
    """The part of the face relative to which the mask should be placed."""

    x_shift: float
    """Shift by X-axis measured in widths of the mask scaled to the face size,
    from left to right. For example, choosing -1.0 will place mask just to the
    left of the default mask position.
    """

    y_shift: float
    """Shift by Y-axis measured in heights of the mask scaled to the face size,
    from top to bottom. For example, 1.0 will place the mask just below the
    default mask position.
    """

    scale: float
    """Mask scaling coefficient. For example, 2.0 means double size."""



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.typing import NoneBool


@final
class DeleteWebhook(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.selete_webhook`"""

    drop_pending_updates: NoneBool = None



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class VideoChatEnded(TelegramType):
    """This object represents a service message about a video chat ended in
    the chat.

    See: https://core.telegram.org/bots/api#videochatended
    """

    duration: int
    """Video chat duration in seconds."""



from __future__ import annotations

__all__ = (
    "UpdateT",
    "UpdateModel",
    "RouterMiddleware",
    "RouterCallable",
    "HandlerCallable",
    "HandlerMiddleware",
    "ReplyMarkup",
    "NoneStr",
    "NoneInt",
    "NoneBool",
)

from collections.abc import Awaitable
from collections.abc import Callable
from typing import TYPE_CHECKING
from typing import Any
from typing import Protocol
from typing import TypeAlias
from typing import TypeVar
from typing import runtime_checkable

if TYPE_CHECKING:
    from .bot import Bot
    from .models import UpdateInfo
    from .types import CallbackQuery
    from .types import ChatBoostRemoved
    from .types import ChatBoostUpdated
    from .types import ChatJoinRequest
    from .types import ChatMemberUpdated
    from .types import ChosenInlineResult
    from .types import ForceReply
    from .types import InlineKeyboardMarkup
    from .types import InlineQuery
    from .types import Message
    from .types import MessageReactionCountUpdated
    from .types import MessageReactionUpdated
    from .types import Poll
    from .types import PollAnswer
    from .types import PreCheckoutQuery
    from .types import ReplyKeyboardMarkup
    from .types import ReplyKeyboardRemove
    from .types import ShippingQuery
    from .types.abc import TelegramType
else:
    TelegramType = Any
    CallbackQuery = Any
    ChatBoostRemoved = Any
    ChatBoostUpdated = Any
    ChatJoinRequest = Any
    ChatMemberUpdated = Any
    ChosenInlineResult = Any
    ForceReply = Any
    InlineKeyboardMarkup = Any
    InlineQuery = Any
    Message = Any
    MessageReactionCountUpdated = Any
    MessageReactionUpdated = Any
    Poll = Any
    PollAnswer = Any
    PreCheckoutQuery = Any
    ReplyKeyboardMarkup = Any
    ReplyKeyboardRemove = Any
    ShippingQuery = Any

T = TypeVar("T")
T_contra = TypeVar("T_contra", contravariant=True)

UpdateT = TypeVar(
    "UpdateT",
    bound=(
        CallbackQuery
        | ChatJoinRequest
        | ChatMemberUpdated
        | InlineQuery
        | ChosenInlineResult
        | Message
        | MessageReactionCountUpdated
        | MessageReactionUpdated
        | Poll
        | PollAnswer
        | PreCheckoutQuery
        | ShippingQuery
        | ChatBoostRemoved
        | ChatBoostUpdated
    )
)

ResultModelT = TypeVar(
    "ResultModelT", bound=TelegramType | list | int | str | bool
)

UpdateModel: TypeAlias = (
    "CallbackQuery "
    "| ChatJoinRequest "
    "| ChatMemberUpdated "
    "| InlineQuery "
    "| ChosenInlineResult "
    "| Message "
    "| MessageReactionCountUpdated "
    "| MessageReactionUpdated "
    "| Poll "
    "| PollAnswer "
    "| PreCheckoutQuery "
    "| ShippingQuery"
    "| ChatBoostRemoved"
    "| ChatBoostUpdated"
)

ReplyMarkup: TypeAlias = (
    "ForceReply "
    "| InlineKeyboardMarkup "
    "| ReplyKeyboardMarkup "
    "| ReplyKeyboardRemove"
)

NoneStr: TypeAlias = "str | None"
NoneInt: TypeAlias = "int | None"
NoneBool: TypeAlias = "bool | None"

GuardCallable: TypeAlias = "Callable[[UpdateT], Awaitable[bool]]"
MiddlewareType: TypeAlias = "Callable[[T], T]"
HandlerCallable: TypeAlias = "Callable[[UpdateT], Awaitable[None]]"
HandlerMiddleware: TypeAlias = "MiddlewareType[HandlerCallable[UpdateT]]"
RouterCallable: TypeAlias = "Callable[[UpdateInfo[UpdateT]], Awaitable[bool]]"
RouterMiddleware: TypeAlias = "MiddlewareType[RouterCallable[UpdateT]]"

PollingHook: TypeAlias = "Callable[[Bot], Awaitable[None]]"


@runtime_checkable
class ErrorHandler(Protocol):
    """Error handler protocol."""

    async def on_error(self, error: Exception) -> None:  # noqa: U100
        """
        :param error: :class:`Exception` instance.
        """
        pass


@runtime_checkable
class Filter(Protocol[T_contra]):
    """Update filter protocol."""

    @property
    def priority(self) -> int:
        """Filter priority. Used for sorting."""
        pass

    def check(self, update: T_contra) -> bool:  # noqa: U100
        """Returns ``True`` if ``update`` matches the filter parameters.

        :param update: Incoming update.
        """
        pass


@runtime_checkable
class InputFile(Protocol):
    """File protocol"""

    @property
    def file_name(self) -> str:
        """Name of file."""
        pass

    async def read(self) -> bytes:
        """Returns the file content."""
        pass



from __future__ import annotations

from typing import final

from msgspec import field

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .order_info import OrderInfo
from .user import User


@final
class PreCheckoutQuery(TelegramType):
    """This object contains information about an incoming pre-checkout query.

    See: https://core.telegram.org/bots/api#precheckoutquery
    """

    id: str
    """Unique query identifier."""

    from_: User = field(name="from")
    """User who sent the query."""

    currency: str
    """Three-letter ISO 4217 currency code."""

    total_amount: int
    """Total price in the smallest units of the currency.

    .. important::

        Integer, not float/double.
    """

    invoice_payload: str
    """Bot specified invoice payload."""

    shipping_option_id: NoneStr = None
    """*Optional.* Identifier of the shipping option chosen by the user."""

    order_info: OrderInfo | None = None
    """*Optional.* Order information provided by the user."""

    async def answer(self, ok: bool, error_message: NoneStr = None) -> bool:
        """See: :meth:`answer_pre_checkout_query <yatbaf.bot.Bot.answer_pre_checkout_query>`"""  # noqa: E501

        return await self.bot.answer_pre_checkout_query(
            pre_checkout_query_id=self.id,
            ok=ok,
            error_message=error_message,
        )



from __future__ import annotations

from typing import final

from yatbaf.enums import MessageEntityType
from yatbaf.typing import NoneStr

from .abc import TelegramType
from .user import User


@final
class MessageEntity(TelegramType):
    """This object represents one special entity in a text message. For example,
    hashtags, usernames, URLs, etc.

    See: https://core.telegram.org/bots/api#messageentity
    """

    type: MessageEntityType
    """Type of the entity."""

    offset: int
    """Offset in UTF-16 code units to the start of the entity."""

    length: int
    """Length of the entity in UTF-16 code units."""

    url: NoneStr = None
    """*Optional.* For :attr:`yatbaf.enums.MessageEntityType.TEXT_LINK` only,
    URL that will be opened after user taps on the text.
    """

    user: User | None = None
    """*Optional.* For :attr:`yatbaf.enums.MessageEntityType.TEXT_MENTION`
    only, the mentioned user.
    """

    language: NoneStr = None
    """*Optional.* For :attr:`yatbaf.enums.MessageEntityType.PRE` only, the
    programming language of the entity text.
    """

    custom_emoji_id: NoneStr = None
    """*Optional.* For :attr:`yatbaf.enums.MessageEntityType.CUSTOM_EMOJI`
    only, unique identifier of the custom emoji. Use
    :meth:`get_custom_emoji_stickers <yatbaf.bot.Bot.get_custom_emoji_stickers>`
    to get full information about the sticker.
    """  # noqa: E501
